package de.noventi;

import de.noventi.util.CryptoUtil;
import de.noventi.util.FileUtil;
import de.noventi.xmlsecpoc.HardcodedEncryptionHelper;
import java.io.File;
import java.nio.charset.Charset;
import java.security.PrivateKey;
import java.security.PublicKey;
import org.apache.commons.cli.CommandLine;
import org.apache.commons.cli.CommandLineParser;
import org.apache.commons.cli.DefaultParser;
import org.apache.commons.cli.HelpFormatter;
import org.apache.commons.cli.Option;
import org.apache.commons.cli.Options;
import org.apache.commons.cli.ParseException;

/**
 * Main class for using the hardcoded encryption / decryption.
 * <p/>
 *
 * @author zeitler
 */
public class Main {

  public static void main(String[] args) {
    Option help = Option.builder("h")
        .longOpt("help")
        .desc("print usage")
        .build();
    Option key = Option.builder("k")
        .numberOfArgs(1)
        .longOpt("key")
        .optionalArg(false)
        .desc("the pkcs12 keystore file containing the key")
        .hasArg()
        .argName("keyfile")
        // .required()
        .build();
    Option input = Option.builder("i")
        .numberOfArgs(1)
        .longOpt("in")
        .optionalArg(false)
        .desc("the input file to read from")
        .hasArg()
        .argName("infile")
        // .required()
        .build();
    Option output = Option.builder("o")
        .numberOfArgs(1)
        .longOpt("out")
        .optionalArg(false)
        .desc("the output file to write to")
        .hasArg()
        .argName("outfile")
        // .required()
        .build();
    Option alias = Option.builder("a")
        .numberOfArgs(1)
        .longOpt("alias")
        .optionalArg(false)
        .desc("the alias of the key within the keystore")
        .hasArg()
        .argName("alias")
        // .required()
        .build();
    Option password = Option.builder("p")
        .numberOfArgs(1)
        .longOpt("password")
        .optionalArg(false)
        .desc("the password to access the keystore and the private key")
        .hasArg()
        .argName("password")
        // .required()
        .build();

    Options options = new Options()
        .addOption(help)
        .addOption(key)
        .addOption(input)
        .addOption(alias)
        .addOption(password)
        .addOption(output);

    try {
      CommandLineParser parser = new DefaultParser();
      CommandLine line = parser.parse(options, args);

      if (line.getArgs().length == 1) {
        if (line.getArgs()[0].equals("encrypt")) {
          checkOptions(line);
          encrypt(line.getOptionValue("i"), line.getOptionValue("k"), line.getOptionValue("a"), line.getOptionValue("p").toCharArray(), line.getOptionValue("o"));
          System.exit(0);
        } else if (line.getArgs()[0].equals("decrypt")) {
          checkOptions(line);
          decrypt(line.getOptionValue("i"), line.getOptionValue("k"), line.getOptionValue("a"), line.getOptionValue("p").toCharArray(), line.getOptionValue("o"));
          System.exit(0);
        }
      }
      printHelp(options);
    }
    catch (ParseException exp) {
      System.err.println( "Parsing failed.  Reason: " + exp.getMessage() );
      printHelp(options);
      System.exit(1);
    }
    catch(Exception e) {
      e.printStackTrace();
      System.exit(2);
    }
  }

  private static void checkOptions(CommandLine line) throws ParseException {
    if (!line.hasOption("i")) {
      throw new ParseException("Missing mandatory option -i");
    }
    if (!line.hasOption("k")) {
      throw new ParseException("Missing mandatory option -k");
    }
    if (!line.hasOption("o")) {
      throw new ParseException("Missing mandatory option -o");
    }
    if (!line.hasOption("a")) {
      throw new ParseException("Missing mandatory option -a");
    }
    if (!line.hasOption("p")) {
      throw new ParseException("Missing mandatory option -p");
    }
  }

  private static void printHelp(Options options) {
    HelpFormatter formatter = new HelpFormatter();
    formatter.printHelp("java -jar xmlsecpoc-all.jar [encrypt|decrypt]", options);
  }

  private static void encrypt(String dataFileName, String keyStoreFileName, String alias, char[] password, String outFileName) {
    File keyStoreFile = FileUtil.getFileForReading(keyStoreFileName);
    File dataFile = FileUtil.getFileForReading(dataFileName);

    try {
      PublicKey publicKey = CryptoUtil.getPublicKey(keyStoreFile, password, alias);
      byte[] data = FileUtil.readFileAsBytes(dataFile);
      byte[] encryptedData = HardcodedEncryptionHelper.encryptData(publicKey, data);

      String encodedData = CryptoUtil.encodeBase64ToString(encryptedData);

      File outputFile = new File(outFileName);
      FileUtil.writeString(outputFile, encodedData);
    } catch (Exception e) {
      throw new IllegalStateException(e);
    }
  }

  private static void decrypt(String dataFileName, String keyStoreFileName, String alias, char[] password, String outFileName) {
    File keyStoreFile = FileUtil.getFileForReading(keyStoreFileName);
    File dataFile = FileUtil.getFileForReading(dataFileName);

    try {
      PrivateKey privateKey = CryptoUtil.getPrivateKey(keyStoreFile, password, alias);
      String dataEncoded = FileUtil.readFileWithEncoding(dataFile, "UTF-8");
      byte[] byteWurst = CryptoUtil.decodeBase64(dataEncoded);
      byte[] bytes = HardcodedEncryptionHelper.decryptData(privateKey, byteWurst);

      String data = new String(bytes, Charset.forName("UTF-8"));
      File outputFile = new File(outFileName);
      FileUtil.writeString(outputFile, data);

    } catch (Exception e) {
      throw new IllegalStateException(e);
    }
  }

}
