package de.noventi.util;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileReader;
import java.io.IOException;
import java.security.KeyFactory;
import java.security.KeyPair;
import java.security.KeyPairGenerator;
import java.security.KeyStore;
import java.security.KeyStoreException;
import java.security.NoSuchAlgorithmException;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.SecureRandom;
import java.security.Security;
import java.security.UnrecoverableKeyException;
import java.security.cert.Certificate;
import java.security.cert.CertificateException;
import java.security.spec.InvalidKeySpecException;
import java.security.spec.PKCS8EncodedKeySpec;
import java.security.spec.X509EncodedKeySpec;
import java.util.Base64;
import javax.crypto.KeyGenerator;
import javax.crypto.SecretKey;
import org.bouncycastle.jce.provider.BouncyCastleProvider;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Class abstract.
 * <p/>
 * Class description
 * <p/>
 *
 * @author zeitler
 */
public class CryptoUtil {

  /**
   * Logger
   */
  private static final Logger LOG = LoggerFactory.getLogger(CryptoUtil.class);

  public static void init() {
    Security.addProvider(new BouncyCastleProvider());
    org.apache.xml.security.Init.init();
    LOG.info("static initializer: crypto initialisation done!");
  }

  public static KeyPair generateRsaKeyPair() throws Exception {
    KeyPairGenerator keyGen = KeyPairGenerator.getInstance("RSA");
    SecureRandom random = SecureRandom.getInstanceStrong();
    keyGen.initialize(4096, random);
    return keyGen.generateKeyPair();
  }

  public static KeyPair getKeyPairFromKeystore(File keyStoreFile, char[] password, String alias) {
    try(FileInputStream fis = new FileInputStream(keyStoreFile)) {
      KeyStore keyStore = KeyStore.getInstance("PKCS12");
      keyStore.load(fis, password);
      PublicKey publicKey = keyStore.getCertificate(alias).getPublicKey();
      PrivateKey privateKey = (PrivateKey) keyStore.getKey(alias, password);
      KeyPair keyPair = new KeyPair(publicKey, privateKey);

      LOG.debug("getKeyPairFromKeystore: retrieved keys from store! store={}, alias={}", keyStoreFile, alias);

      return keyPair;
    } catch (KeyStoreException | IOException | NoSuchAlgorithmException | CertificateException | UnrecoverableKeyException e) {
      throw new IllegalStateException(e);
    }
  }

  public static Certificate getCertificate(File keyStoreFile, char[] password, String alias) {
    try {
      KeyStore keyStore = loadKeyStore(keyStoreFile, password);
      Certificate certificate = keyStore.getCertificate(alias);

      LOG.debug("getCertificate: retrieved certificate from store! store={}, alias={}", keyStoreFile, alias);

      return certificate;
    } catch (KeyStoreException e) {
      throw new IllegalStateException(e);
    }
  }

  public static PublicKey getPublicKey(File keyStoreFile, char[] password, String alias) {
    try {
      KeyStore keyStore = loadKeyStore(keyStoreFile, password);
      PublicKey publicKey = keyStore.getCertificate(alias).getPublicKey();

      LOG.debug("getPublicKey: retrieved public key from store! store={}, alias={}", keyStoreFile, alias);

      return publicKey;
    } catch (KeyStoreException e) {
      throw new IllegalStateException(e);
    }
  }
  
  public static PrivateKey getPrivateKey(File keyStoreFile, char[] password, String alias) {
    try {
      KeyStore keyStore = loadKeyStore(keyStoreFile, password);
      PrivateKey privateKey = (PrivateKey) keyStore.getKey(alias, password);

      LOG.debug("getPrivateKey: retrieved private key from store! store={}, alias={}", keyStoreFile, alias);

      return privateKey;
    } catch (KeyStoreException | UnrecoverableKeyException | NoSuchAlgorithmException e) {
      throw new IllegalStateException(e);
    }
  }

  public static KeyStore loadKeyStore(File keyStoreFile, char[] password) {
    try(FileInputStream fis = new FileInputStream(keyStoreFile)) {
      KeyStore keyStore = KeyStore.getInstance("PKCS12");
      keyStore.load(fis, password);
      
      LOG.debug("loadKeyStore: loaded Key-Store. keyStoreFile={}", keyStoreFile);
      
      return keyStore;
    } catch (KeyStoreException | NoSuchAlgorithmException | CertificateException | IOException e) {
      throw new IllegalStateException(e);
    }
  }
  
  public static SecretKey generateDataEncryptionKey() throws Exception {
    return generateDataEncryptionKey("AES", 256);
  }

  public static SecretKey generateDataEncryptionKey(String algorithm, int keyLenght) throws Exception {
    KeyGenerator keyGenerator = KeyGenerator.getInstance(algorithm);
    keyGenerator.init(keyLenght);
    return keyGenerator.generateKey();
  }

  public static String encodeBase64(String data) {
    return Base64.getEncoder().encodeToString(data.getBytes());
  }

  public static String encodeBase64ToString(byte[] data) {
    return Base64.getEncoder().encodeToString(data);
  }

  public static String decodeBase64ToString(String data) {
    return new String(Base64.getDecoder().decode(data));
  }

  public static byte[] decodeBase64(String data) {
    return Base64.getDecoder().decode(data);
  }

  public static PrivateKey getPrivateKey(File file)  {
    String privateKeyPEM = getKey(file);
    return getPrivateKeyFromString(privateKeyPEM);
  }

  public static String getPrivateKeyEncoded(File file) {
    PrivateKey privateKey = getPrivateKey(file);
    return org.apache.commons.codec.binary.Base64.encodeBase64String(privateKey.getEncoded());
  }

  public static PrivateKey getPrivateKeyFromString(String key) {
    try {
      String privateKeyPEM = key;
      privateKeyPEM = privateKeyPEM.replace("-----BEGIN PRIVATE KEY-----\n", "");
      privateKeyPEM = privateKeyPEM.replace("-----END PRIVATE KEY-----", "");
      byte[] encoded = org.apache.commons.codec.binary.Base64.decodeBase64(privateKeyPEM);
      KeyFactory kf = KeyFactory.getInstance("RSA");
      PKCS8EncodedKeySpec keySpec = new PKCS8EncodedKeySpec(encoded);
      PrivateKey privKey = kf.generatePrivate(keySpec);
      return privKey;
    } catch (NoSuchAlgorithmException | InvalidKeySpecException e) {
      throw new IllegalStateException(e);
    }
  }


  public static PublicKey getPublicKey(File file) {
    String publicKeyPEM = getKey(file);
    return getPublicKeyFromString(publicKeyPEM);
  }

  public static String getPublicKeyEncoded(File file) {
    PublicKey privateKey = getPublicKey(file);
    return org.apache.commons.codec.binary.Base64.encodeBase64String(privateKey.getEncoded());
  }

  public static PublicKey getPublicKeyFromString(String key) {
    try {
      String publicKeyPEM = key;
      publicKeyPEM = publicKeyPEM.replace("-----BEGIN PUBLIC KEY-----\n", "");
      publicKeyPEM = publicKeyPEM.replace("-----END PUBLIC KEY-----", "");
      byte[] encoded = org.apache.commons.codec.binary.Base64.decodeBase64(publicKeyPEM);
      KeyFactory kf = KeyFactory.getInstance("RSA");
      PublicKey pubKey = kf.generatePublic(new X509EncodedKeySpec(encoded));
      return pubKey;
    } catch (NoSuchAlgorithmException | InvalidKeySpecException e) {
      throw new IllegalStateException(e);
    }
  }

  private static String getKey(File file) {
    try (BufferedReader br = new BufferedReader(new FileReader(file))) {
      String strKeyPEM = "";
      String line;
      while ((line = br.readLine()) != null) {
        strKeyPEM += line + "\n";
      }
      return strKeyPEM;
    } catch (IOException e) {
      throw new IllegalStateException(e);
    }
  }

}
