package de.noventi.xmlsecpoc;

import de.noventi.util.CryptoUtil;
import java.nio.ByteBuffer;
import java.security.PrivateKey;
import java.security.PublicKey;
import java.security.SecureRandom;
import java.security.spec.AlgorithmParameterSpec;
import java.util.Arrays;
import javax.crypto.Cipher;
import javax.crypto.SecretKey;
import javax.crypto.spec.IvParameterSpec;
import javax.crypto.spec.SecretKeySpec;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Helper for hardcoded encrypting / decrypting data providing an IV and data encryption key.
 * <p/>
 *
 * @author zeitler
 */
public class HardcodedEncryptionHelper {

  /**
   * Logger
   */
  private static final Logger LOG = LoggerFactory.getLogger(HardcodedEncryptionHelper.class);

  private static final String DATA_ENCRYPTION_KEY_ALG = "AES";

  private static final int DATA_ENCRYPTION_KEY_LENGHT = 256;

  private static final int DATA_ENCRYPTION_IV_LENGHT = 16;

  private static final int ENCRYPTED_KEY_LENGHT = 32;

  // Used for compatibility with provided encrypted data example
  private static final String DATA_ENCRYPTION_ALG = "AES/CBC/PKCS5Padding";
  // private static final String DATA_ENCRYPTION_ALG = "AES/GCM/NoPadding";

  private static final String KEY_ENCRYPTION_ALG = "RSA/ECB/OAEPWithSHA1AndMGF1Padding";

  private HardcodedEncryptionHelper() {
  }

  /**
   * Encrypts a bunch of data.
   * Key and IV are generated by a strong secure random.
   * No validation etc. done.
   * @param publicKey the public key for encrypting the generated secret key.
   * @param data the data for encryption.
   * @return the contatinated data, containing the encrypted secret key, the iv and the encrypted data.
   * @throws Exception lazy throwing the world...
   */
  public static byte[] encryptData(PublicKey publicKey, byte[] data) throws Exception {
    SecureRandom secureRandom = SecureRandom.getInstanceStrong();
    // create key for data encryption
    SecretKey secretKey = CryptoUtil.generateDataEncryptionKey(DATA_ENCRYPTION_KEY_ALG, DATA_ENCRYPTION_KEY_LENGHT);
    byte[] secretKeyEncoded = secretKey.getEncoded();

    // create IV
    byte[] iv = new byte[DATA_ENCRYPTION_IV_LENGHT];
    secureRandom.nextBytes(iv);

    // encrypt data with created key
    Cipher cipher = Cipher.getInstance(DATA_ENCRYPTION_ALG);
    AlgorithmParameterSpec parameterSpec = new IvParameterSpec(iv);
    cipher.init(Cipher.ENCRYPT_MODE, secretKey, parameterSpec);
    byte[] encryptedData = cipher.doFinal(data);

    // concatinate the secret key and the iv
    ByteBuffer buffer = ByteBuffer.allocate(secretKeyEncoded.length + iv.length);
    byte[] secretKeyAndIv = buffer.put(secretKeyEncoded).put(iv).array();

    // encrypt secretKeyAndIv with public key
    cipher = Cipher.getInstance(KEY_ENCRYPTION_ALG);
    cipher.init(Cipher.ENCRYPT_MODE, publicKey);
    byte[] encryptedKeyAndIv = cipher.doFinal(secretKeyAndIv);

    // concat bytes: encrypted key and iv + encrypted data
    buffer = ByteBuffer.allocate(encryptedKeyAndIv.length + encryptedData.length);
    buffer.put(encryptedKeyAndIv).put(encryptedData);

    return buffer.array();
  }

  /**
   * Decrypt some data containing an encrypted secret key and an IV along with the encrypted data.
   * No validation etc. done.
   * @param privateKey the private key for decrypting the secret key.
   * @param byteWurst the concatinated data, containing the encrypted secret key, the IV and the encrypted data.
   * @return the decrypted data.
   * @throws Exception lazy throwing the world...
   */
  public static byte[] decryptData(PrivateKey privateKey, byte[] byteWurst) throws Exception {
    // cut data chunk into single components
    byte[] encryptedKeyAndIv = Arrays.copyOfRange(byteWurst, 0, DATA_ENCRYPTION_KEY_LENGHT);
    byte[] encryptedData = Arrays.copyOfRange(byteWurst, DATA_ENCRYPTION_KEY_LENGHT, byteWurst.length);

    // decrypt the secret key and iv with the private key
    Cipher cipher = Cipher.getInstance(KEY_ENCRYPTION_ALG);
    cipher.init(Cipher.DECRYPT_MODE, privateKey);
    byte[] secretKeyAndIv = cipher.doFinal(encryptedKeyAndIv);

    // extract secret key and iv
    byte[] secretKeyEncoded = Arrays.copyOfRange(secretKeyAndIv, 0, ENCRYPTED_KEY_LENGHT);
    byte[] iv = Arrays.copyOfRange(secretKeyAndIv, ENCRYPTED_KEY_LENGHT, ENCRYPTED_KEY_LENGHT + DATA_ENCRYPTION_IV_LENGHT);

    // decrypt the data with the secret key
    SecretKey secretKey = new SecretKeySpec(secretKeyEncoded, 0, secretKeyEncoded.length, DATA_ENCRYPTION_KEY_ALG);
    cipher = Cipher.getInstance(DATA_ENCRYPTION_ALG);
    AlgorithmParameterSpec parameterSpec = new IvParameterSpec(iv);
    cipher.init(Cipher.DECRYPT_MODE, secretKey, parameterSpec);

    return cipher.doFinal(encryptedData);
  }

}
