package de.noventi.xmlsecpoc;

import de.noventi.util.CryptoUtil;
import java.security.InvalidKeyException;
import java.security.Key;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.PublicKey;
import java.security.SignatureException;
import java.security.cert.CertificateException;
import java.security.cert.X509Certificate;
import java.util.Collections;
import java.util.Date;
import java.util.Iterator;
import java.util.List;
import javax.xml.crypto.AlgorithmMethod;
import javax.xml.crypto.KeySelector;
import javax.xml.crypto.KeySelectorException;
import javax.xml.crypto.KeySelectorResult;
import javax.xml.crypto.XMLCryptoContext;
import javax.xml.crypto.XMLStructure;
import javax.xml.crypto.dsig.CanonicalizationMethod;
import javax.xml.crypto.dsig.Reference;
import javax.xml.crypto.dsig.SignedInfo;
import javax.xml.crypto.dsig.Transform;
import javax.xml.crypto.dsig.XMLSignature;
import javax.xml.crypto.dsig.XMLSignatureFactory;
import javax.xml.crypto.dsig.dom.DOMSignContext;
import javax.xml.crypto.dsig.dom.DOMValidateContext;
import javax.xml.crypto.dsig.keyinfo.KeyInfoFactory;
import javax.xml.crypto.dsig.keyinfo.X509Data;
import javax.xml.crypto.dsig.spec.C14NMethodParameterSpec;
import javax.xml.crypto.dsig.spec.TransformParameterSpec;
import org.apache.xml.security.encryption.EncryptedData;
import org.apache.xml.security.encryption.EncryptedKey;
import org.apache.xml.security.encryption.XMLCipher;
import org.apache.xml.security.keys.KeyInfo;
import org.apache.xml.security.utils.EncryptionConstants;
import org.bouncycastle.util.encoders.Base64;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/**
 * Helper for doing XML security.
 * <p/>
 * Sign, encrypt, decrypt, verify.
 *
 * @author zeitler
 */
public class XmlWrappingEncryptionHelper {

  /**
   * Logger
   */
  private static final Logger LOG = LoggerFactory.getLogger(XmlWrappingEncryptionHelper.class);

  private static final String KEY_ENCRYPTION_ALGORITHM = "http://www.w3.org/2001/04/xmlenc#rsa-oaep-mgf1p";

  // The targeted .NET versions are not able to use other algorithms then CBC.
  private static final String CONTENXT_ENCRYPTION_ALGORITHM = "http://www.w3.org/2001/04/xmlenc#aes256-cbc";
  // private static final String CONTENXT_ENCRYPTION_ALGORITHM = "http://www.w3.org/2009/xmlenc11#aes256-gcm";

  private static String XML_SIGNATURE_ALGORITHM = "http://www.w3.org/2001/04/xmldsig-more#rsa-sha512";

  private static String XML_DIGEST_METHOD = "http://www.w3.org/2001/04/xmlenc#sha512";


  private XmlWrappingEncryptionHelper() {
  }

  /**
   * Encrypt a XML document with a provided key encryption key. The date encryption key in
   * generated on the fly.
   * @param document the XML document.
   * @param keyEncryptionKey the key encryption key.
   * @throws Exception lazy throwing the world...
   */
  public static void encrypt(Document document, Key keyEncryptionKey) throws Exception {
    Key dataEncryptionKey = CryptoUtil.generateDataEncryptionKey();

    XMLCipher keyCipher = XMLCipher.getInstance(KEY_ENCRYPTION_ALGORITHM);
    keyCipher.init(XMLCipher.WRAP_MODE, keyEncryptionKey);

    EncryptedKey encryptedKey = keyCipher.encryptKey(document, dataEncryptionKey);

    Element element = document.getDocumentElement(); // root
    // example to use just a single tag instead of the XML root element
    // element = getDirectChild(element, "EVerordnung");

    XMLCipher xmlCipher = XMLCipher.getInstance(CONTENXT_ENCRYPTION_ALGORITHM);
    xmlCipher.init(XMLCipher.ENCRYPT_MODE, dataEncryptionKey);

    EncryptedData encryptedData = xmlCipher.getEncryptedData();

    KeyInfo keyInfo = new KeyInfo(document);
    keyInfo.add(encryptedKey);
    encryptedData.setKeyInfo(keyInfo);

    xmlCipher.doFinal(document, element, true);
  }

  /**
   * Decrypting a XML document.
   * @param document the XML document.
   * @param keyDecryptionKey the key decryption key.
   * @throws Exception lazy throwing the world...
   */
  public static void decrypt(Document document, Key keyDecryptionKey) throws Exception {
    Element encryptedDataElement = (Element) document.getElementsByTagNameNS(
        EncryptionConstants.EncryptionSpecNS,
        EncryptionConstants._TAG_ENCRYPTEDDATA).item(0); // the one and only

    XMLCipher xmlCipher = XMLCipher.getInstance();
    xmlCipher.init(XMLCipher.DECRYPT_MODE, null);
    xmlCipher.setKEK(keyDecryptionKey);
    xmlCipher.doFinal(document, encryptedDataElement);
  }

  /**
   * Sign (enveloped) a XML document embedding the X509 certificate for signature validation.
   * @param document the XML documemnt.
   * @param signatureKey the signing key
   * @param certificate the matching certificate.
   * @throws Exception lazy throwing the world...
   */
  public static void sign(Document document, Key signatureKey, X509Certificate certificate) throws Exception {
    document.normalizeDocument();

    XMLSignatureFactory fac = XMLSignatureFactory.getInstance("DOM");

    Reference ref = fac.newReference(
        "", // todo: URI?
        fac.newDigestMethod(XML_DIGEST_METHOD, null),
        Collections
            .singletonList(fac.newTransform(Transform.ENVELOPED, (TransformParameterSpec) null)),
        null,
        null
    );

    SignedInfo signedInfo = fac.newSignedInfo(
        fac.newCanonicalizationMethod(
            CanonicalizationMethod.EXCLUSIVE, (C14NMethodParameterSpec) null),
        fac.newSignatureMethod(XML_SIGNATURE_ALGORITHM, null),
        Collections.singletonList(ref));

    KeyInfoFactory kif = fac.getKeyInfoFactory();
    X509Data x509d = kif.newX509Data(Collections.singletonList(certificate));
    javax.xml.crypto.dsig.keyinfo.KeyInfo signingKeyInfo = kif.newKeyInfo(Collections.singletonList(x509d));

    DOMSignContext dsc = new DOMSignContext(signatureKey, document.getDocumentElement());

    XMLSignature signature = fac.newXMLSignature(signedInfo, signingKeyInfo);

    signature.sign(dsc);

    LOG.info("sign: signed documment! document={}", document);
  }

  /**
   * Verify a XML signature using the public key withtin the provided X509 certificate.
   * @param sourceDoc the XML document
   * @throws Exception lazy throwing the world...
   */
  public static void verifySignature(Document sourceDoc) throws Exception {
    verifySignature(sourceDoc, null);
  }

  /**
   * Verify a XML signature using the provided X509 data but comparing against a provided public key
   * if provided. The acutal date is used to check certificate validity. This is just for testing
   * purpose. In real world, the validity of the certificate must be checked, e.g. against time
   * of signing, the issuer, a CRL, ...
   * @param sourceDoc the XML document.
   * @param publicKey the key, may be <code>null</code>
   * @throws Exception lazy throwing the world...
   */
  public static void verifySignature(Document sourceDoc, PublicKey publicKey) throws Exception {
    sourceDoc.normalizeDocument();

    KeySelector keySelector = new SubmittedX509KeySelector(publicKey);

    NodeList nl = sourceDoc.getElementsByTagNameNS(javax.xml.crypto.dsig.XMLSignature.XMLNS, "Signature");
    XMLSignatureFactory fac = XMLSignatureFactory.getInstance("DOM");
    DOMValidateContext valContext = new DOMValidateContext(keySelector, nl.item(0));

    XMLSignature signature = fac.unmarshalXMLSignature(valContext);
    String signingAlgorithm = signature.getSignedInfo().getSignatureMethod().getAlgorithm().intern();
    String messageDigestAlgorithm = ( (Reference) signature.getSignedInfo().getReferences().get(0) ).getDigestMethod().getAlgorithm().intern();

    LOG.info("verifySignature: verifying signature! siginingAlgorithm={}, messageDigestAlgorithm={}", signingAlgorithm, messageDigestAlgorithm);

    boolean coreValidity = signature.validate(valContext);

    if (!coreValidity) {
      // optional, provide some info on invalid core validity
      boolean sv = signature.getSignatureValue().validate(valContext);
      LOG.warn("verifySignature: signature valid={}", sv);

      // check validation state
      Iterator i = signature.getSignedInfo().getReferences().iterator();
      for (int j = 0; i.hasNext(); j++) {
        Reference reference = ( (Reference) i.next() );
        boolean refValid = reference.validate(valContext);
        LOG.warn("verifySignature: reference {} verifikations state: {}", j, refValid);
        LOG.warn("verifySignature: digestValue: {}", new String(Base64.encode(reference.getCalculatedDigestValue())));
        LOG.warn("verifySignature: digestMethod: {}", reference.getDigestMethod().toString());
      }

      throw new IllegalStateException("Invalid signature!");
    }
  }

  /**
   * Key selector for verification using a provided public key and the actual date to
   * check validity, just to check somethiing at all.
   */
  private static class SubmittedX509KeySelector extends KeySelector {

    private PublicKey publicKey;

    private SubmittedX509KeySelector(PublicKey publicKey) {
      this.publicKey = publicKey;
    }

    public KeySelectorResult select(javax.xml.crypto.dsig.keyinfo.KeyInfo keyInfo,
        Purpose purpose, AlgorithmMethod method,
        XMLCryptoContext context) throws KeySelectorException {

      if (keyInfo == null) {
        throw new KeySelectorException("Null KeyInfo object!");
      }
      List list = keyInfo.getContent();

      XMLStructure xmlStructure = (XMLStructure) list.get(0); // one... or none
      if (xmlStructure instanceof X509Data) {
        X509Data x509 = (X509Data) xmlStructure;
        for (Object content : x509.getContent()) {
          if (content instanceof X509Certificate) {
            X509Certificate certificate = (X509Certificate) content;

            Date disposalDate = new Date(); // fake news!
            verifyAndValidateCertificate(certificate, publicKey, disposalDate);

            PublicKey pk = certificate.getPublicKey();
            return new SimpleKeySelectorResult(pk);
          }
        }
      }

      throw new KeySelectorException("No KeyValue element found!");
    }

    private void verifyAndValidateCertificate(X509Certificate certificate, PublicKey publicKey, Date disposalDate) {
      // do whatever needs to be done to trust the provided certificate
      try {
        certificate.checkValidity(disposalDate);
        if (publicKey != null) { // don't do this in real life!
          certificate.verify(publicKey);
          LOG.info("verifyAndValidateCertificate: verified against known public key and disposal date!");
        } else {
          LOG.info("verifyAndValidateCertificate: verified against disposal date!");
        }

      } catch (CertificateException e) {
        throwVerifiyingOrValidationException("Error in certificate!");
      } catch (NoSuchAlgorithmException e) {
        throwVerifiyingOrValidationException("Invalid algorightm!");
      } catch (InvalidKeyException e) {
        throwVerifiyingOrValidationException("Invalid key!");
      } catch (NoSuchProviderException e) {
        throwVerifiyingOrValidationException("Invalid provider!");
      } catch (SignatureException e) {
        throwVerifiyingOrValidationException("Invalid signature!");
      }
    }

    private void throwVerifiyingOrValidationException(String message) {
      LOG.warn("throwVerifiyingOrValidationException: message={}", message);
      throw new IllegalStateException(message);
    }
  }

  /**
   * Key Selector result.
   */
  private static class SimpleKeySelectorResult implements KeySelectorResult {

    private PublicKey pk;

    SimpleKeySelectorResult(PublicKey pk) {
      this.pk = pk;
    }

    public Key getKey() {
      return pk;
    }
  }

}
