package de.noventi.xmlsecpoc;

import static org.assertj.core.api.AssertionsForClassTypes.assertThatThrownBy;
import static org.assertj.core.api.AssertionsForInterfaceTypes.assertThat;

import de.noventi.util.CryptoUtil;
import de.noventi.util.DomUtil;
import de.noventi.util.FileUtil;
import java.io.File;
import java.security.KeyPair;
import java.security.cert.X509Certificate;

import org.junit.BeforeClass;
import org.junit.Test;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.w3c.dom.Document;

/**
 * Class abstract.
 * <p/>
 * Class description
 * <p/>
 *
 * @author zeitler
 */
public class TestXmlEncryptionHelper {

  /**
   * Logger
   */
  private static final Logger LOG = LoggerFactory.getLogger(TestXmlEncryptionHelper.class);

  private static final File outputDir = new File("./build/out/xml");

  private static final File resourceDir = new File("./src/test/resources");

  @BeforeClass
  public static void beforeClass() throws Exception {
    CryptoUtil.init();

    // Produces an encrypted XML file using the Java default, which is prefixed namespaces.
    // XmlWrappingEncryptionHelper.init(XmlWrappingEncryptionHelper.DefaultNamespaceHandling.NONE);

    // Produces an encrypted XML file using xenc as default namespace and ds as prefixed namespace.
    // XmlWrappingEncryptionHelper.init(XmlWrappingEncryptionHelper.DefaultNamespaceHandling.XENC);

    // Produces a .NET compliant encrypted XML file using several default namespaces.
    XmlWrappingEncryptionHelper.init(XmlWrappingEncryptionHelper.DefaultNamespaceHandling.EXPLICIT_DEFAULTS);

    if (outputDir.exists()) {
      FileUtil.deleteDirectory(outputDir.toPath());
    }
    outputDir.mkdirs();
  }

  @Test
  public void testWrappedEncryptionDecryptionWithDomWithoutSigning() throws Exception {
    File encryptionKeyStore = new File(resourceDir, "pharmatechnik/xml/cert_and_key.p12");
    assertThat(encryptionKeyStore).isNotNull().exists();
    KeyPair encryptionKeyPair = CryptoUtil.getKeyPairFromKeystore(encryptionKeyStore, "123456".toCharArray(), "selfsigned");

    Document document = DomUtil.readXmlFromFile(new File(resourceDir, "pharmatechnik/xml/eDispensierung.xml"), true);
    DomUtil.docToFile(document, new File(outputDir, "eDispensierung.xml").getAbsolutePath());

    // encrypt
    XmlWrappingEncryptionHelper.encrypt(document, encryptionKeyPair.getPublic());
    DomUtil.docToFile(document, new File(outputDir, "eDispensierung-encrypted-unsigned.xml").getAbsolutePath());

    // read as new document to do a new marshalling
    Document readDocument = DomUtil.readXmlFromFile(new File(outputDir, "eDispensierung-encrypted-unsigned.xml"), true);

    // decrypt
    XmlWrappingEncryptionHelper.decrypt(readDocument, encryptionKeyPair.getPrivate());
    DomUtil.docToFile(readDocument, new File(outputDir, "eDispensierung-decrypted-unsigned.xml").getAbsolutePath());

    //<eDispensierung><Transaktionsnummer>123456789</Transaktionsnummer>
    assertThat(readDocument.getDocumentElement().getElementsByTagName("Transaktionsnummer")).isNotNull();
  }


  @Test
  public void testWrappedEncryptionDecryptionWithDom() throws Exception {
    File encryptionKeyStore = new File(resourceDir, "pharmatechnik/xml/cert_and_key.p12");
    assertThat(encryptionKeyStore).isNotNull().exists();
    KeyPair encryptionKeyPair = CryptoUtil.getKeyPairFromKeystore(encryptionKeyStore, "123456".toCharArray(), "selfsigned");

    File signingKeyStore = new File(resourceDir, "signing_cert_and_key.p12");
    assertThat(signingKeyStore).isNotNull().exists();
    KeyPair signingKeyPair = CryptoUtil.getKeyPairFromKeystore(signingKeyStore, "123456".toCharArray(), "serverkeys");
    X509Certificate signingCertificate = (X509Certificate) CryptoUtil.getCertificate(signingKeyStore, "123456".toCharArray(), "serverkeys");

    Document document = DomUtil.readXmlFromFile(new File(resourceDir, "pharmatechnik/xml/eDispensierung.xml"), true);
    DomUtil.docToFile(document, new File(outputDir, "eDispensierung.xml").getAbsolutePath());

    // sign and encrypt
    XmlWrappingEncryptionHelper.sign(document, signingKeyPair.getPrivate(), signingCertificate);
    DomUtil.docToFile(document, new File(outputDir, "eDispensierung-signed.xml").getAbsolutePath());

    XmlWrappingEncryptionHelper.encrypt(document, encryptionKeyPair.getPublic());

    DomUtil.docToFile(document, new File(outputDir, "eDispensierung-encrypted.xml").getAbsolutePath());

    // read as new document to do a new marshalling
    Document readDocument = DomUtil.readXmlFromFile(new File(outputDir, "eDispensierung-encrypted.xml"), true);

    // decrypt and verify
    XmlWrappingEncryptionHelper.decrypt(readDocument, encryptionKeyPair.getPrivate());
    DomUtil.docToFile(readDocument, new File(outputDir, "eDispensierung-decrypted.xml").getAbsolutePath());

    // verify with known public key
    XmlWrappingEncryptionHelper.verifySignature(readDocument, signingKeyPair.getPublic());
    // verify against provided public key from signature certificate
    XmlWrappingEncryptionHelper.verifySignature(readDocument);

    // failing singature verification due to wrong public key
    assertThatThrownBy(()
        -> XmlWrappingEncryptionHelper.verifySignature(readDocument, encryptionKeyPair.getPublic())
    ).isInstanceOf(IllegalStateException.class);

    //<eDispensierung><Transaktionsnummer>123456789</Transaktionsnummer>
    assertThat(readDocument.getDocumentElement().getElementsByTagName("Transaktionsnummer")).isNotNull();
  }

  @Test
  public void testDecryptionFromCSharp() throws Exception {
    File keyStore = new File(resourceDir, "pharmatechnik/xml/cert_and_key.p12");
    KeyPair keyPair = CryptoUtil.getKeyPairFromKeystore(keyStore, "123456".toCharArray(), "selfsigned");

    Document document = DomUtil.readXmlFromFile(new File(resourceDir, "pharmatechnik/xml/eDispensierung_signiert_encrypted.xml"), true);

    XmlWrappingEncryptionHelper.decrypt(document, keyPair.getPrivate());
    XmlWrappingEncryptionHelper.verifySignature(document, keyPair.getPublic());

    DomUtil.docToFile(document, new File(outputDir, "eDispensierung-csharp-decrypted.xml").getAbsolutePath());

    //<eDispensierung><Transaktionsnummer>123456789</Transaktionsnummer>
    assertThat(document.getDocumentElement().getElementsByTagName("Transaktionsnummer")).isNotNull();
  }

}
